/*
 * (C) 2005 by Bengt Martensson <nospam@bengt-martensson.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <iostream>
#include <stdlib.h>
#include <string.h>

static bool debug_assumptions = false;
static bool debug_print_first_table = false;
static bool debug_print_all_tables = false;
static bool debug_print_no_trues = false;
static bool count_solutions = false;

#define rule(nr, type, prop1, value1, prop2, value2) \
   void rule##nr(table &t) { type(t, prop1, value1, prop2, value2);}
#define rule_three(nr, type, prop1, value1, prop2, value2, prop3, value3) \
   void rule##nr(table &t) { type(t, prop1, value1, prop2, value2, prop3, value3);}
#define rule_four(nr, type, prop1, value1, prop2, value2, prop3, value3, prop4, value4) \
   void rule##nr(table &t) { type(t, prop1, value1, prop2, value2, prop3, value3, prop4, value4);}

enum truth_value { Undetermined = -1, False = 0, True = 1 };

class table
{
public:
  char name[20];
  int property;
  bool array[no_entries][no_properties][no_entries];
  table();
  table(char *, int);
  table(table *);
  void print();
  int no_trues();
  truth_value satisfied();
  const char * satisfied_text();
};

typedef void rulef(table &);

// Global data
rulef *rules[no_rules];

truth_value
table::satisfied()
{
  bool unique = true;
  for (int i = 0; i < no_entries; i++)
    for (int j = 0; j < no_properties; j++) {
      int no_t = 0;
      for (int k = 0; k < no_entries; k++) {
	if (array[i][j][k])
	  no_t++;
      }
      if (no_t == 0)
	return False;
      if (no_t > 1)
	unique = false;
    }
  return unique ? True : Undetermined;
}

const char *
table::satisfied_text()
{
  return satisfied() == True ? "TRUE" : satisfied() == False
    ? "FALSE" : "Undetermined";
}

void
prettyprint(table *t[no_properties], bool print_table)
{
  /*static*/ bool virgin = true;
  if (virgin && t[0]->satisfied() == True) {
    virgin = false;
    if (print_table)
      t[0]->print();
    for (int row = 0; row < no_properties; row++) {
      std::cout << t[row]->name << "|";
      for (int column = 0; column < no_entries; column++) {
	for (int k = 0; k < no_entries; k++) {
	  if (t[0]->array[column][row][k])
	    std::cout << " " << words[row][k];
	}
      }
      std::cout << "\n";
    }
  }
}

table::table(table *original)
{
  //name = (char*) malloc(strlen(original->name) + 1);
  strcpy(name, original->name);
  property = original->property;
  for (int i = 0; i < no_entries; i++)
    for (int j = 0; j < no_properties; j++)
      for (int k = 0; k < no_entries; k++)
	array[i][j][k] = original->array[i][j][k];
}

int
table::no_trues()
{
  int s = 0;
  for (int i = 0; i < no_entries; i++)
    for (int j = 0; j < no_properties; j++)
      for (int k = 0; k < no_entries; k++)
	if (array[i][j][k])
	  s++;
  return s;
}

#if 0
table::table()
{
  for (int i = 0; i < no_entries; i++) {
    for (int j = 0; j < no_properties; j++) {
      for (int k = 0; k < no_entries; k++) {
	//cout << i << j << k << "\n";
	array[i][j][k] = true;
      }
    }
  }
}
#endif

table::table(char * nam, int init_prop)
{
  //name = (char*) malloc(strlen(nam) + 1);
  strcpy(name, nam);
  property = init_prop;
  for (int i = 0; i < no_entries; i++) {
    for (int j = 0; j < no_properties; j++) {
      for (int k = 0; k < no_entries; k++) {
	//cout << i << j << k << "\n";
	array[i][j][k] = (j == init_prop) ? (k == i) : true;
      }
    }
  }
}

void
table::print()
{
  std::cout << name << " (" << satisfied_text() << ")\n  ";
  for (int i = 0; i < no_properties; i++) {
    std::cout << property_words[i] << "    ";
    for (int n = 9; n < no_entries; n++)
      std::cout << " ";
  }
  
  std::cout << "\n-------------------------------------------------------------------------\n";
      
  for (int i = 0; i < no_entries; i++) { // One line
    std::cout << i << ": ";
    for (int j = 0; j < no_properties; j++) {
      for (int k = 0; k < no_entries; k++) {
	std::cout << (array[i][j][k] ? "t" : " ");
	//	std::cout << i << j << k << (array[i][j][k] ? "t" : "f") << "\n";
      }
      for (int n = no_entries - 1; n < 8; n++)
	std::cout << " ";
      std::cout << " | ";
    }
      std::cout << "\n";
  }
}

// For a row of a table, does this row say that properto called property
// surely has the value value?
bool
fact(bool ent[no_properties][no_entries], int property, int value)
{
  int no_t = 0;
  for (int i = 0; i < no_entries; i++)
    if (ent[property][i])
      no_t++;
  //std::cout << property << value << no_t << ent[property][value] << "YES!\n";
  return (no_t == 1) && ent[property][value];
}

bool
antifact(bool ent[no_properties][no_entries], int property, int value)
{
  return ! ent[property][value];
}

void
setantifact(bool ent[no_properties][no_entries], int property, int value)
{
  for (int i = 0; i < no_entries; i++) {
    //std::cout << property << value << "YESsds!\n";
    ent[property][value] = false;
  }
}

void
setfact(bool ent[no_properties][no_entries], int property, int value)
{
  for (int i = 0; i < no_entries; i++) {
    //std::cout << property << value << "YESsds!\n";
    ent[property][i] = (i == value);
  }
}

void
equivalence1(table &t, int prop1, int value1, int prop2, int value2)
{
  for (int i = 0; i < no_entries; i++) {
    if (fact(t.array[i], prop1, value1)) {
      setfact(t.array[i], prop2, value2);
    }
    if (antifact(t.array[i], prop1, value1)) {
      setantifact(t.array[i], prop2, value2);
    }
  }
}
    
void
equivalence2(table &t, int prop1, int value1, int prop2, int value2)
{
  equivalence1(t, prop1, value1, prop2, value2);
  equivalence1(t, prop2, value2, prop1, value1);
}
    
#ifdef HAVE_POSITION
void
neighbor1(table &t, int prop1, int value1, int prop2, int value2)
{
  if (t.property == position) {
    //std::cout << prop1 << value1 << prop2 << value2 << "\n";
    for (int i = 0; i < no_entries; i++) {
      if (fact(t.array[i], prop1, value1)) {
	for (int j = 0; j < i - 1; j++)
	  setantifact(t.array[j], prop2, value2);
	setantifact(t.array[i], prop2, value2);
	//std::cout << "\n    " << i << prop2 << value2 << "\n";
	for (int j = i + 2; j < no_entries; j++)
	  setantifact(t.array[j], prop2, value2);
      }
    }
  }
}
    
void
neighbor2(table &t, int prop1, int value1, int prop2, int value2)
{
  if (t.property == position) {
    neighbor1(t, prop1, value1, prop2, value2);
    neighbor1(t, prop2, value2, prop1, value1);
  }
}
    
void
to_right2(table &t, int prop1, int value1, int prop2, int value2)
{
  if (t.property == position) {
    t.array[0][prop1][value1] = false;
    t.array[no_entries-1][prop2][value2] = false;
    for (int i = 0; i < no_entries - 1; i++) {
      if (!(t.array[i+1][prop1][value1] && t.array[i][prop2][value2])) {
	t.array[i+1][prop1][value1] = false;
	t.array[i][prop2][value2] = false;
      }
    }
  }
}

void
facing2(table &t, int prop1, int value1, int prop2, int value2)
{
  int dist = no_entries/2;
  if (t.property == position) {
    for (int i = 0; i < no_entries; i++) {
      if (fact(t.array[i], prop1, value1))
	setfact(t.array[(i + dist) % no_entries], prop2, value2);
      if (fact(t.array[i], prop2, value2))
	setfact(t.array[(i + dist) % no_entries], prop1, value1);

      if (!(t.array[(i+dist)%no_entries][prop1][value1] && t.array[i][prop2][value2])) {
      t.array[(i+dist)%no_entries][prop1][value1] = false;
      t.array[i][prop2][value2] = false;
      }
    }
  }
}

void
not_neighbor2(table &t, int prop1, int value1, int prop2, int value2)
{
  if (t.property == position)
    for (int i = 0; i < no_entries; i++)
      if (fact(t.array[i], prop1, value1))
	for (int j = i-1; j <= i+1; j++)
	  if ((j >= 0) && (j < no_entries))
	    setantifact(t.array[j], prop2, value2);
}

void
gegenueber2(table &t, int prop1, int value1, int prop2, int value2)
{
  const int gegen[] = { 6, 5, 4, 7, 2, 1, 0, 3};
  if (t.property == position) {
    for (int i = 0; i < 8; i++) {
      if (!(t.array[i][prop1][value1] && t.array[gegen[i]][prop2][value2])) {
	t.array[i][prop1][value1] = false;
	t.array[gegen[i]][prop2][value2] = false;
      }
    }
  }
}
#endif // HAVE_POSITION


void
exclusion2(table &t, int prop1, int value1, int prop2, int value2)
{
  for (int i = 0; i < no_entries; i++) {
    if (fact(t.array[i], prop1, value1)) {
      setantifact(t.array[i], prop2, value2);
    }
    if (fact(t.array[i], prop2, value2)) {
      setantifact(t.array[i], prop1, value1);
    }
  }
}

void
exclusion3(table &t, int prop1, int value1, int prop2, int value2,
	   int prop3, int value3)
{
  exclusion2(t, prop1, value1, prop2, value2);
  exclusion2(t, prop1, value1, prop3, value3);
  exclusion2(t, prop2, value2, prop3, value3);
}

void
exclusion4(table &t, int prop1, int value1, int prop2, int value2,
	   int prop3, int value3, int prop4, int value4)
{
  exclusion2(t, prop1, value1, prop2, value2);
  exclusion2(t, prop1, value1, prop3, value3);
  exclusion2(t, prop1, value1, prop4, value4);
  exclusion2(t, prop2, value2, prop3, value3);
  exclusion2(t, prop2, value2, prop4, value4);
  exclusion2(t, prop3, value3, prop4, value4);
}

void
property_in2(table &t, int prop1, int value1, int prop2, const bool value_table[no_entries])
{
#if 1
  for (int i = 0; i < no_entries; i++)
    if (fact(t.array[i], prop1, value1))
      for (int j = 0; j < no_entries; j++)
	t.array[i][prop2][j] &= value_table[j];

  if (t.property == prop2)
    for (int i = 0; i < no_entries; i++)
      t.array[i][prop1][value1] &= value_table[i];
#else
  for (int i = 0; i < no_entries; i++) 
    if (fact(t.array[i], prop1, value1)) {
      for (int j = 0; j < no_entries; j++) 
	if (!value_table[j]) {
	  //std::cout << "prohibit " << prop1 << "\t" << value1 << words[prop1][value1] << " from " << prop2 << "\t" << j << words[prop2][j]<< "\n";
	  setantifact(t.array[i], prop2, j);
	}
    }
#endif
}

#if 0
void
rule_trivial(table &t)
{
}
#endif

void
use_permutation_prop(table &t)
{
  for (int p = 0; p < no_properties; p++) {
    int index = -1;
    for (int e = 0; e < no_entries; e++) {
      int no_t = 0;
      for (int i = 0; i < no_entries; i++) {
	if (t.array[i][p][e]) {
	  no_t++;
	  index = i;
	}
      }
      if (no_t == 1 && t.property != p) {
	//std::cout << "Table " << t.property << ": Property " << p << ", value " << e << " pos " << index << " is unique\n";
      //else
      //std::cout << no_t << "\n";
	for (int i = 0; i < no_entries; i++) {
	  //t.array[e][p][i] = i == e;
	  t.array[index][p][i] = i == e;
	}
      }
    }

    for (int e = 0; e < no_entries; e++) {
      int no_t = 0;
      for (int i = 0; i < no_entries; i++) {
	if (t.array[e][p][i]) {
	  no_t++;
	  index = i;
	}
      }
      if (no_t == 1 && t.property != p) {
	//std::cout << "Table " << t.property << ": Property " << p << ", value " << e << " is unique\n";
	//else
	//std::cout << no_t << "\n";
	for (int i = 0; i < no_entries; i++) {
	  t.array[i][p][index] = i == e;
	}
      }
    }
  }
}

int
no_trues(table *tables[no_properties])
{
  int s = 0;
  for (int table_index = 0; table_index < no_properties; table_index++)
    s += tables[table_index]->no_trues();
  return s;
}

truth_value
deduce(table *tables[no_properties])
{
  //std::cout << "entering deduce\n";
  int no_t;
  do {
    if (debug_print_no_trues)
      std::cout << "==>" << no_trues(tables) << "\n";
    no_t = no_trues(tables);
    for (int the_table = 0; the_table < no_properties; the_table++) {

      // Apply all the rules
      for (int i = 0; i < no_rules; i++)
	rules[i](*tables[the_table]);

      // See if some property/value-pair are unique
      use_permutation_prop(*tables[the_table]);
    }

    // Transfer information between tables
    for (int the_table = 0; the_table < no_properties; the_table++) {
      int prop = tables[the_table]->property;
      // For each property...
      for (int table_index = 0; table_index < no_properties; table_index++) {
	if (tables[table_index]->property != prop) {
	  //... search the other tables if they contain sure information
	  for (int i = 0; i < no_entries; i++) {
	    // for each row in these tables, 
	    // does it have property prop unique?
	    int no_t = 0;
	    int val = -1;
	    for (int j = 0; j < no_entries; j++) {
	      if (tables[table_index]->array[i][prop][j]) {
		no_t++;
		val = j;
	      }
	    }
	    if (no_t == 1) {
	      // yes, it had, its value was val
	      // i.e., tables
	      //std::cout << table_index << prop << val << "\n";
	      // Merge this information to both tables
	      for (int p = 0; p < no_properties; p++) {
		for (int v = 0; v < no_entries; v++) {
		  bool b = 
		    tables[the_table]->array[val][p][v] && 
		    tables[table_index]->array[i][p][v];
		  tables[the_table]->array[val][p][v] = b;
		  tables[table_index]->array[i][p][v] = b;
		}
	      }
	    }
	  }
	}
      }
    }
  } while (no_t != no_trues(tables));

  //std::cout << tables[0]->satisfied();

  if (debug_print_first_table || debug_print_all_tables)
    tables[0]->print();
  if (debug_print_all_tables)
    for (int the_table = 1; the_table < no_properties; the_table++)
      tables[the_table]->print();
  bool determined = true;
  for (int i = 0; i < no_properties; i++) {
    switch (tables[i]->satisfied()) {
    case False:
      return False;
      break;
    case True:
      break;
    case Undetermined:
      determined = false;
      break;
    }
  }
  return determined ? True : Undetermined;    
}

const char *
levelpadder(int level)
{
  const char indents = 2;
  static char s[1000];
  for (int i = 0; i < indents*level; i++)
    s[i] = ' ';
  s[indents*level] = '\0';
  return s;
}

int number_solutions = 0;

bool
process(table *tables[no_properties], int level)
{
  truth_value result = deduce(tables);
  bool success = result == True;
  if (result == Undetermined) {
    // Must make an assumption and call myself recursively.
    // for this, find first undetermined property/value pair
    int prop = -1;
    int row = -1;
    for (int i = 0; i < no_entries && prop == -1; i++) {
      for (int p = 0; p < no_properties && prop == -1; p++) {
	int no_t = 0;
	for (int j = 0; j < no_entries; j++) {
	  if (tables[0]->array[i][p][j])
	    no_t++;
	}
	if (no_t > 1) {
	  prop = p;
	  row = i;
	}
      }
    }
    
    //std::cout << prop << row << "\n";
    bool found_one_solution = false;
    for (int i = 0; i < no_entries; i++) {
      // Try with letting property prop in row row have value i
      if (tables[0]->array[row][prop][i]) {
	// Need a new copy of the working tables
	table *tabs[no_properties];
	for (int j = 0; j < no_properties; j++) {
	  tabs[j] = new table(tables[j]);
	}

	if (debug_assumptions) {
	  std::cout << levelpadder(level) << "[" << level << "] "
	       << (level < 10 ? " " : "") << "Row " << row
	       << ": Assuming property "
	       << prop << " has value " << i << "\n";
	}
	for (int k = 0; k < no_entries; k++) {
	  tabs[0]->array[row][prop][k] = k == i;
	}
	//tabs[0]->print();
	success = process(tabs, level+1);
	if (success) {
	  if (debug_assumptions) {
	    std::cout << levelpadder(level) << "[" << level << "] "
		 << (level < 10 ? " " : "") << "TRUE: Row " << row
		 << ": Assuming property " << prop << " has value " << i
		 << "\n";
	  }
	  prettyprint(tabs, true);
	  if (tabs[0]->satisfied() == True) {
	    number_solutions++;
	    found_one_solution = true;
	  }
	  //break;
	  if (!count_solutions)	{
	    for (int j = 0; j < no_properties; j++)
	      delete tabs[j];

	    return true;
	  }
	}
	if (! success && debug_assumptions) {
	  std::cout << levelpadder(level) << "[" << level << "] "
	       << (level < 10 ? " " : "") << "FALSE: Row " << row
	       << ": Assuming property " << prop << " has value " << i
	       << "\n";
	}
	for (int j = 0; j < no_properties; j++)
	  delete tabs[j];
      }
    }
    success = found_one_solution;
  }
  if (success && level == 0)
    prettyprint(tables, true);
  return success;
}

void
decode_debugarg(const char *str)
{
  debug_assumptions = bool(strchr(str, 'a'));
  debug_print_first_table = bool(strchr(str, 'f'));
  debug_print_all_tables = bool(strchr(str, 't'));
  debug_print_no_trues = bool(strchr(str, 'n'));
  count_solutions = bool(strchr(str, '#'));
}
